import numpy as np
from scipy.special import erfc
from ase.data import covalent_radii, atomic_numbers
from ase.calculators.calculator import Calculator, all_changes
from ase.neighborlist import NeighborList
from ions.data import principle_number
from ions.potential import BVFF
import numpy as np
from scipy.spatial.distance import cdist, euclidean


class BVSECalculator(Calculator):
    """
    BVSE calculator based on ase Calculator. It is used for NEB calculations
    of a mobile ion migration activation energy. 
    """

    implemented_properties = ["energy", "forces"]
    default_parameters = {
        'rc': 10.0,
    }
    nolabel = True # check

    def __init__(self, site = None, nl = None, **kwargs):
        """

        Args:
            site: index of site in ase.Atoms for which NEB, int
                Note: Considering source, taret notation it is the source
            nl: ase NeighborList, optional
                object should have method get_neighbors(site)
            **kwargs:
        """
        super().__init__(**kwargs)
        self.site = site
        self.nl = nl

    def calculate(
        self,
        atoms = None,
        properties = None,
        system_changes = None,
    ):
        """
        Args:
            atoms (ase.Atoms): ase Atoms object
            properties (list): list of properties to calculate
            system_changes (list): monitor which properties of atoms were
                changed for new calculation. If not, the previous calculation
                results will be loaded.
        Returns:

        """
        properties = properties or ["energy"]
        system_changes = system_changes or all_changes
        super().calculate(atoms=atoms, properties=properties, system_changes=system_changes)
        
        natoms = len(self.atoms)
        rc = self.parameters.rc

        if self.nl is None or 'numbers' in system_changes:
            self.nl = NeighborList(
                [rc / 2] * natoms, self_interaction=False, bothways=True, skin = 0.3
            )

        self.nl.update(self.atoms)

        energies = np.zeros(natoms)
        forces = np.zeros((natoms, 3))
        positions = self.atoms.positions
        cell = self.atoms.cell
        neighbors, offsets = self.nl.get_neighbors(self.site)
        cells = np.dot(offsets, cell)
        r = self.atoms.positions[self.site] - (atoms.positions[neighbors] + np.dot(atoms.cell.T, offsets.T).T)
        r2 = np.linalg.norm(r, axis = 1)

        q1 = self.atoms.get_array('oxi_states')[self.site]
        s1 = self.atoms.symbols[self.site]
        z1 = self.atoms.numbers[self.site]
        n1 = principle_number[z1]
        rc1 = covalent_radii[z1]
        q2 = np.take(atoms.get_array('oxi_states'), neighbors)
        q2 = np.where(q2 * q1 < 0.0, 0, q2)
        n2 = np.take(atoms.get_array('n2'), neighbors)
        rc2 = np.take(atoms.get_array('rc2'), neighbors)
        alpha =  np.take(atoms.get_array('alpha'), neighbors)
        r_min =  np.take(atoms.get_array('r_min'), neighbors)
        d0 =  np.take(atoms.get_array('d0'), neighbors)
        mask = np.take(atoms.get_array('mask'), neighbors)
        
        e1 = BVFF.Morse(r2, r_min, d0, alpha).sum() 
        e2 = (mask * BVFF.Coulomb(q1, q2, r2, rc1, rc2, n1, n2)).sum()
        e = e1 + e2
        
        energies[self.site] = e 
        energy = energies.sum()

        x, y, z = r[:, 0], r[:, 1], r[:, 2]
        dfdx1 = BVFF.dMorse_dX(x, alpha, r_min, d0, r2).sum()
        dfdy1 = BVFF.dMorse_dX(y, alpha, r_min, d0, r2).sum()
        dfdz1 = BVFF.dMorse_dX(z, alpha, r_min, d0, r2).sum()

        dfdx2 = (mask * BVFF.dCoulomb_dX(x, q1, q2, rc1, rc2, n1, n2, r2)).sum()
        dfdy2 = (mask * BVFF.dCoulomb_dX(y, q1, q2, rc1, rc2, n1, n2, r2)).sum()
        dfdz2 = (mask * BVFF.dCoulomb_dX(z, q1, q2, rc1, rc2, n1, n2, r2)).sum()

        grad = np.array([dfdx1 + dfdx2, dfdy1 + dfdy2, dfdz1 + dfdz2])
        forces[self.site] = -grad
        self.results['energy'] = energy
        self.results['energies'] = energies
        self.results['free_energy'] = energy
        self.results['forces'] = forces



    
class BVFFCalculator(Calculator):
    """
    Not tested.
    """

    implemented_properties = ["energy", "forces"]
    default_parameters = {
        'rc': 10.0,
    }
    nolabel = True # What is?

    def __init__(self, V, site = None, nl = None, **kwargs):
        """

        Args:
            site: index of site in ase.Atoms for which NEB, int
                Note: Considering source, taret notation it is the source
            nl: ase NeighborList, optional
                object should have method get_neighbors(site)
            **kwargs:
        """
        super().__init__(**kwargs)
        self.V = V
        self.size = np.array(V.shape)
        self.mesh = np.argwhere(V) / (self.size - 1)
        self.site = site
        self.nl = nl

    def locate(self, p):

        X, Y, Z = self.mesh[:, 0], self.mesh[:, 1], self.mesh[:, 2]
        d, idx = self.closest_pythagoras(X, Y, Z, p)
        if d > 0.2:
            raise
        return idx


    def closest_pythagoras(self, X, Y, Z, P):
        """
            X: x-axis series
            Y: y-axis series
            Z: z-axis series
            P: centre point to measure distance from
            
            Returns: tuple(closest point, distance between P and closest point)
        """ 
        # Compute the distances between each of the points and the desired point using Pythagoras' theorem
        distances = np.sqrt((X - P[0]) ** 2 + (Y - P[1]) ** 2 + (Z - P[2]) ** 2)
        # Get the index of the smallest value in the array to determine which point is closest and return it
        idx_of_min = np.argmin(distances)
        #(np.array([X[idx_of_min], Y[idx_of_min], Z[idx_of_min]])
        return distances[idx_of_min], idx_of_min



    def calculate(
        self,
        atoms = None,
        properties = None,
        system_changes = None,
    ):
        """
        Args:
            atoms (ase.Atoms): ase Atoms object
            properties (list): list of properties to calculate
            system_changes (list): monitor which properties of atoms were
                changed for new calculation. If not, the previous calculation
                results will be loaded.
        Returns:

        """
        properties = properties or ["energy"]
        system_changes = system_changes or all_changes
        super().calculate(atoms=atoms, properties=properties, system_changes=system_changes)
        natoms = len(atoms)
        energies = np.zeros(natoms)
        forces = np.zeros((natoms, 3))

        p = atoms.get_scaled_positions()[self.site]
        idx = self.locate(p)
        e = self.V.reshape(-1)[idx]

        energies[self.site] = e 
        energy = energies.sum()


        dr = self.atoms.cell.cellpar()[:3] / self.size
        gradV = np.gradient(self.V, edge_order=2)  # how to treat pbc?
        dfdx = - gradV[0].reshape(-1)[idx]/dr[0]
        dfdy = - gradV[1].reshape(-1)[idx]/dr[1]
        dfdz = - gradV[1].reshape(-1)[idx]/dr[2]

        forces[self.site] = np.array([dfdx, dfdy, dfdz])
        self.results['energy'] = energy
        self.results['energies'] = energies
        self.results['free_energy'] = energy
        self.results['forces'] = forces



import itertools
import os
import json
import numpy as np
import networkx as nx
from scipy.spatial import cKDTree
from ase import Atoms
from ase.io import write, read
from ase.neighborlist import NeighborList
from ase.neb import NEB
from ase.spacegroup import get_spacegroup
from ase.neighborlist import neighbor_list
from ase.build import make_supercell
from ase.data import covalent_radii
from spglib import get_symmetry_dataset
from spglib import standardize_cell
from ions.tools import SaddleFinder



__version__ = "0.1"

        
class PathFinder:
    
    """ 
    Perconeb object.
        
    The class can be used to find the percolating pathways of a mobile specie in a framework. 
    The functionality includes: 
    
    - calculating 1-3D percolation radius for a given mobile specie in a structure
    - searching for a minimum cutoff for maximum jump distance of a mobile specie required for 1-3D percolation
    - finding percolation pathway and its inequivalent parts (tests are required)

    For more details read the docs. 
    
    """ 
    def __init__(self, atoms, specie: int, upper_bound: float, symprec = 1e-3, oxi_states = False):
        
        """
        
        Parameters
        ----------

        atoms: ase's Atoms object
            Should contain a mobile specie of interest

        specie: int
            atomic number of a mobile specie, e.g. 11 for Na

        upper_bound: float, 10.0 by default
            maximum jump distance between equilibrium sites of a mobile specie
            
        symprec: float, 1e-3 by default
            precision for a space group analysis

        oxi_states: boolean, False by default
            whether atoms has 'oxi_states' attribute
            
        """
        
        
        self.specie = specie
        self.symprec = symprec
        self.upper_bound = min(atoms.cell.cellpar()[:3].max(), upper_bound) + 0.1
        self.oxi_states = oxi_states
        self._set_symmetry_labels(atoms)
        if self.oxi_states:
            self._set_ionic_radii(atoms)
        self.atoms = atoms.copy()
        self.mobile_atoms = self.atoms[self.atoms.numbers == specie]
        self.freezed_atoms = self.atoms[self.atoms.numbers != specie]
        self.mobile_atoms.set_array('unitcell_idx', np.argwhere(atoms.numbers == specie).ravel())
    
    
    
    # def _set_ionic_radii(self, atoms):

    #     symbols = atoms.symbols
    #     charges = atoms.get_array('oxi_states')
    #     r_i = np.array([ionic_radii[s][o] for (s, o) in zip(symbols, charges)])
    #     atoms.set_array('r_i', r_i)
        
        
        
    def _set_symmetry_labels(self, atoms):
        
        spglib_cell = (atoms.cell,
                       atoms.get_scaled_positions(),
                       atoms.numbers)
        equivalent_sites = get_symmetry_dataset(spglib_cell, symprec=self.symprec)['equivalent_atoms']
        atoms.set_array('sym_label', np.array(equivalent_sites))
        
    
    
    def _lineseg_dists(self, p, a, b):
        
        # source:    https://stackoverflow.com/questions/54442057/
        # calculate-the-euclidian-distance-between-an-array-of-points-to-a-line-segment-in/
        # 54442561#54442561 
        
        if np.all(a == b):
            return np.linalg.norm(p - a, axis=1)
        d = np.divide(b - a, np.linalg.norm(b - a))
        s = np.dot(a - p, d)
        t = np.dot(p - b, d)
        h = np.maximum.reduce([s, t, np.zeros(len(p))])
        c = np.cross(p - a, d)
        return np.hypot(h, np.linalg.norm(c, axis = 1))
        
        
        
    def _collect_edges_within_supercell(self):
        
        mobile_atoms = self.mobile_atoms.copy()
        n_vertices = len(mobile_atoms)
        scale = [
            [2, 0, 0],
            [0, 2, 0],
            [0, 0, 2]
        ]
        
        supercell = make_supercell(mobile_atoms.copy(), scale) 
        supercell.pbc = False # we are interested in the peroclation within the supercell
        shifts = np.where((supercell.get_scaled_positions() * 2.0).round(4) >= 1.0, 1, 0)
        supercell.set_array('shift', shifts)
        self.mobile_supercell = supercell
        i, j, d = neighbor_list('ijd', supercell, self.upper_bound)
        ij = np.array(list(zip(i, j)))
        if len(ij) == 0:
            raise
        ij.sort(axis = 1) # (source, target)  == (target, source)
        pairs, idx = np.unique(ij, axis = 0, return_index=True) # remove duplicates
        offsets = np.squeeze(np.diff(supercell.get_array('shift')[pairs], axis = 1))
        self.pairs = pairs
        self.offsets = offsets
        unwrapped_edges = np.hstack([pairs, offsets])
        wrapped_edges = np.hstack([pairs - n_vertices * (pairs // n_vertices), offsets])

        return unwrapped_edges, wrapped_edges, d[idx]
    


    def _annotate_edges(self):

        u, w, jump_distances = self._collect_edges_within_supercell()
        unique_edges, ue_idx, inverse = np.unique(w, axis = 0, return_index = True, return_inverse = True)
        unitcell_idx = self.mobile_atoms.get_array('unitcell_idx')
        distances = []
        for edge  in w[ue_idx]:
            offset = edge[2:]
            source = unitcell_idx[int(edge[0])]
            #source = edge[0]
            #target = edge[1]
            target = unitcell_idx[int(edge[1])]
            shift = np.where(offset < 0, 1, 0)
            p1 = self.atoms.positions[source] + np.dot(shift, self.atoms.cell)
            p2 = self.atoms.positions[target] + np.dot(offset + shift, self.atoms.cell)
            base = self.atoms[[i for i in range(len(self.atoms)) if i not in [source, target]]]
            translations = np.array(list(itertools.product(
                                            [0, 1, -1], # should be [0, 1, -1, 2, -2] ideally
                                            [0, 1, -1],
                                            [0, 1, -1])))
            coords = []
            idx = []
            for tr in translations:
                coords.append(base.positions + np.dot(tr, base.cell))
                idx.append(np.arange(0, len(base)))
            ii = np.hstack(idx)
            p = np.vstack(coords)
            dd = self._lineseg_dists(p, p1, p2)
            d_min = min(dd)
            if self.oxi_states:
                d_min = dd.min() - max(base.get_array('r_i')[ii[dd == dd.min()]])
            distances.append(d_min)
        self.distances = np.take(distances, inverse)
        self.jump_distances = jump_distances
        self.u = u
        self.w = w


    def _filter_edges(self, tr = 0.5, cutoff = 1e2):

        accept = []
        for i, (u, jump, dist) in enumerate(zip(self.u, self.jump_distances, self.distances)):
            if (jump > cutoff) or (dist < tr):
                continue
            else:
                accept.append(i)
        if len(accept) > 1:
            accept = np.array(accept)
        else:
            accept = []
        return accept
                
        
        
    def _percolation_dimensionality(self, edgelist):
        
        n_species = len(self.mobile_atoms)       
        G = nx.from_edgelist(edgelist)
        _, idx = np.unique(self.mobile_atoms.get_array('sym_label'),
                                   return_index = True) # index should be from self.mobile_atoms
        perco_dims_per_site = {}
        sym_uniq_sites = np.arange(0, n_species)[idx]
        for i in sym_uniq_sites:
            dim = 0
            for j in range(0, 2 ** 3):
                i_next_cell = i + j * n_species
                
                try:
                    path_idx = nx.algorithms.shortest_path(G, i, i_next_cell)
                    dim += 1
                    perco_dims_per_site.update({i:dim})
                except:
                    continue
        return perco_dims_per_site
        
        
                
    def percolation_threshold(self, dim, cutoff = 10.0):

        """
        Calculates maximum distance between percolating edges
        and the framework sublattice. 
        If self.oxi_states = True the percolation ionic radius is calculated.
        
        
        Parameters
        ----------

        dim: int, 2 by default
            percolation dimensionality 2 -> 1D, 4 -> 2D, 8 -> 3D

        cutoff: float, 10.0 by default
            maximum allowed jump distance between sites in the mobile sublattice
        """
    
        self._annotate_edges()
        
        emin = 0 # legacy naming
        emax = 10.0
        tr = 0
        while (emax - emin) > 0.01:
            probe = (emin + emax) / 2
            mask = self._filter_edges(tr = probe)
            edges = self.u[mask, :2]
            if len(edges) > 0:
                try:
                    data = self._percolation_dimensionality(edges)
                    if max(list(data.values())) >= dim:
                        emin = probe
                        tr = round(emin,4)
                    else:
                        emax = probe
                except:
                    emax = probe
            else:
                emax = probe
        return tr

    
    
    def cutoff_search(self, dim = 2, tr = 0.75):
        
        """
        Calculates minimum value of a jump distance between sites in a mobile sublattice
        required to form a 1-3D percolating network.
        
        
        Parameters
        ----------

        dim: int, 2 by default
            percolation dimensionality 2 -> 1D, 4 -> 2D, 8 -> 3D

        tr: float, 0.75 by default
            percolation threshold for an edge, i.e. minium distance between edge (line segment)
            between two sites in the mobile sublattice and the framework sublattie below which 
            edge is rejected

        upper_bound: float, 10.0 by default
            starting cutoff value for the search algorithm
        """
        
        self._annotate_edges()

        emin = 0.0 # legacy naming
        emax = self.upper_bound
        
        cutoff = -1.0
        while (emax - emin) > 0.01:
            probe = (emin + emax) / 2
            mask = self._filter_edges(tr = tr, cutoff = probe)
            edges = self.u[mask, :2]
            if len(edges) > 0:
                try:
                    data = self._percolation_dimensionality(edges)
                    if max(list(data.values())) >= dim:
                        emax = probe
                        cutoff = emax
                    else:
                        emin = probe
                except:
                    emin = probe
            else:
                emin = probe
        return cutoff



    def unique_edges(self, tr = 0.75, dim = 2, cutoff = False):
        
        if cutoff:
            mask = self._filter_edges(tr = tr, cutoff = cutoff)
        else:
            cutoff = self.cutoff_search(dim = dim, tr = tr)
            mask = self._filter_edges(tr = tr, cutoff = cutoff)
        self.accepted_mask = mask
        s = np.vstack(self.mobile_atoms.get_array('sym_label')[self.w[:, :2]][mask])
        s.sort(axis = 1)
        d = self.distances[mask].round(3)
        j = self.jump_distances[mask].round(3)
        unique_pairs, idx, inverse = np.unique(np.column_stack((s, d, j)), axis = 0,
                                              return_index = True, return_inverse = True)
        return self.w[mask][idx], idx, inverse
    


    def percolation_dimensionality(self, tr = 0.75):
        max_dim = 0
        dim_cutoff = 0.0
        for dim in [2, 4, 8]:
            cutoff = self.cutoff_search(dim = dim, tr = tr)
            if cutoff > 0.0:
                max_dim = dim
                dim_cutoff = cutoff
        return max_dim, dim_cutoff

    


    def propagate_barriers(self, dim, percoedges, emins, emaxs):

        """
        Used for postprocessing. Defines percolation barriers for the 
        passed percolating edges and associated with them min and max 
        energies in the energy profile.
        
        
        Parameters
        ----------

        dim: int, 2 by default
            percolation dimensionality 2 -> 1D, 4 -> 2D, 8 -> 3D

        percoedges: list of N x [source, target], where N is the number of edges
            unwrapped percolating edges

        emins: list of N x float
            minimum energies in the energy profile for each edge

        emaxs: list of N x float
            maximum energies in the energy profile for each edge

        """

        #self._annotate_edges()
        emin = min(emins)
        emax = max(emaxs)
        tr = False
        unique_energies = np.unique(emaxs)
        unique_energies.sort()
        for e in unique_energies[::-1]:
            probe = e
            mask = emaxs <= probe
            edges = percoedges[mask]
            if len(edges) > 0:
                try:
                    data = self._percolation_dimensionality(edges)
                    if max(list(data.values())) >= dim:
                        emin = probe
                        tr = emin
                    else:
                        emax = probe
                except:
                    emax = probe
            else:
                emax = probe

        if tr:
            tr2 = False
            unique_energies = np.unique(emins)
            unique_energies.sort()
            mask = emaxs <= tr
            percoedges = percoedges[mask]
            emax = tr
            for e in unique_energies:
                probe = e
                mask2 = emins[mask] >= probe
                edges = percoedges[mask2]
                if len(edges) > 0:
                    try:
                        data = self._percolation_dimensionality(edges)
                        if max(list(data.values())) >= dim:
                            emin = probe
                            tr2 = emin
                        else:
                            emin = probe
                    except:
                        emin = probe
                else:
                    emin = probe
        else:
            print('Unexpected behavior')
            raise
        
        return abs(tr2 - tr), tr2, tr




class Perconeb(PathFinder):

    def __init__(self, atoms, specie, upper_bound, self_interaction = True):

        self.sf = SaddleFinder(self_interaction=self_interaction)
        super().__init__(atoms, specie, upper_bound)



    def mincut_maxdim(self, tr = 0.75):
        dim, cutoff = self.percolation_dimensionality(tr = tr)
        return dim, cutoff
    

    def percolating_jumps(self, tr = 0.75, min_sep_dist = 10.0, spacing = 0.5):

        dim, cutoff = self.mincut_maxdim(tr = tr)
        ue, ids, inverse = self.unique_edges(cutoff = cutoff, dim = dim, tr = tr)
        edges = self.w[self.accepted_mask]

        traj = []
        for i, edge in enumerate(ue):
            idx = self.mobile_atoms.get_array('unitcell_idx')
            source, target = idx[edge[:2]]
            offset = edge[2:]
            images = self.sf.interpolate(self.atoms, source, target, offset,
                                          spacing = spacing,
                                          min_sep_dist = min_sep_dist
                                        )
            traj.append(images)
        return traj
    


    def create_neb(self, images, k = 5.0, method = 'default', gm = True, **kwargs):
        _, neb = self.sf.bvse_neb(images, k = k, method = method, **kwargs)
        return neb
    


import numpy as np
from ase.neb import NEB
from ions.utils import collect_bvse_params
from ions.calculators import BVSECalculator
from ions.utils import geometric_median_with_bounds
from ase.build import make_supercell
from scipy.spatial import cKDTree



class SaddleFinder:
    """  
    Nudged Elastic band implementation for bond valence force field.
    """
    def __init__(self, self_interaction = True):

        """  
        Initialization

        Parameters
        ----------

        self_interaction: boolean, True by default
            whether to consider interaction between mobile species

        """
        self.self_interaction = self_interaction

    
    def interpolate(self, atoms, source, target, offset, min_sep_dist = 10.0, spacing = 0.5):
        """
        Linearly interpolates trajectory between source and target ions. 
        Uses min_sep_dist to create supercell. 

        Parameters
        ----------

        atoms: Ase's atoms object
            structure should contain mobile species of interest

        source: int
            index of the source positions

        target: int
            index of the target positions

        offset: list of int
            unitcell boundary crossing offset vector between source and target, e.g. [1, 0, 0]
        
        min_sep_dist: float, 10.0 by default
            minimum separation distance between moving ion and its periodic replica
        
        spacing: float, 0.5 by default
            length of the linear segments for the linear interpolation
            Note: if the number of images in the iterpolated trajectory is > 9,
                  spacing will be automatically recalculated for 9 images.

        Returns
        ----------
        images, list of atoms objects
            linearly interpolated trajectory
        """
        symbol = atoms.symbols[source]
        charge = int(atoms.get_array('oxi_states')[source])
        collect_bvse_params(atoms, symbol, charge, self_interaction = self.self_interaction)

        scale = np.ceil(min_sep_dist/atoms.cell.cellpar()[:3]).astype(int)
        p1 = atoms.positions[source]
        p2 = atoms.positions[target] + np.dot(offset, atoms.cell)
        
        d = np.linalg.norm(p1 - p2)
        n_images = int(d // spacing)
        if n_images % 2 == 0:
            n_images += 1
        if n_images > 9:
            n_images = 9

        P = [
            [scale[0], 0, 0],
            [0, scale[1], 0],
            [0, 0, scale[2]]
        ]
        supercell = make_supercell(atoms.copy(), P)
        scaled_edge = supercell.cell.scaled_positions([p1, p2]).round(10) # rounding for a safer wrapping
        scaled_edge[:]%=1.0 #wrapping
        wrapped_edge = supercell.cell.cartesian_positions(scaled_edge)
        if np.linalg.norm(wrapped_edge[0] - wrapped_edge[1]) < 0.1:
            print('source == target', source, target)
            print('min_sep_dist is too small for this jump', source, target, offset)
            raise
        tree = cKDTree(supercell.positions)
        dd, ii = tree.query(wrapped_edge)
        if dd.max() > 1e-6:
            print(dd.max())
            raise

        source_new = ii[0]
        target_new = ii[1]
        assert supercell.symbols[source_new] == symbol
        assert supercell.symbols[target_new] == symbol
        assert source_new != target_new
        traj = np.linspace(p1, p2, n_images)
        images = []
        freezed = np.array([i != source_new for i in range(len(supercell))])
        supercell.set_array('freezed', freezed)
        for p in traj:
            image = supercell.copy()
            image.positions[source_new] = p
            image = image[[i for i in range(len(image)) if i != target_new]]
            image = image[image.numbers.argsort()]
            #image.wrap()
            images.append(image)
            assert len(image) == len(supercell) - 1
        return images



    def bvse_neb(self, images, k = 5.0, default = True, gm = True, **kwargs):

        """Wrapper for ase's NEB object. Sets BVSECalculator for each image.

        Parameters
        ----------

        images: list of Ase's atoms object
            structure should contain mobile species of interest
            and array named "freezed" defined as
            freezed = np.array([i != source for i in range(len(supercell))])
            atoms.sett_array('freezed', freezed)
            see .interpolate method for the details

        k: float, 5.0 by default
            string force costant, eV
        
        default: boolean, whether to use default method or not, True by default
            if False, will use **kwargs passed for ase's NEB object

        gm: boolean, True by default
            shift moving ion in the center image into its geometric median within supercell
        
        """

        site = np.argwhere(images[0].get_array('freezed') == False).ravel()[0]
        for image in images:
            image.calc =  BVSECalculator(site = site)

        if gm:
            center = images[len(images) // 2].copy()
            left = images[len(images) // 2 - 1].copy()
            right = images[len(images) // 2 + 1].copy()
            gm, y0 = geometric_median_with_bounds(left, center, right, site)
            images[len(images) // 2].positions[site] = np.array(gm)

        if default:
            neb = NEB(
                    images,
                    climb=False,
                    k = k,
                    method = 'improvedtangent',
                    )
        else:
            neb = NEB(images, k = k, **kwargs)
        return images, neb
    


    def get_barrier(self, images):
        """returns barrier for the passed images"""
        e = []
        for image in images:
            e.append(image.get_potential_energy())
        barrier = max(e) - min(e)
        return barrier
    

    def get_profile(self, images):
        """returns energy profile for the passed images"""
        profile = []
        for image in images:
            profile.append(image.get_potential_energy())
        return np.array(profile)
    

    def get_forces(self, images):
        """returns forces acting on moving ion for the passed images"""
        forces = []
        for image in images:
            site = np.argwhere(image.get_array('freezed') == False).ravel()[0]
            forces.append(image.get_forces()[site])
        return np.array(forces)


import itertools
import numpy as np
from scipy.spatial.distance import cdist, euclidean
from scipy.optimize import minimize
from ase.data import atomic_numbers, covalent_radii
from ions.data import bvse_data, principle_number

def collect_bvse_params(atoms, symbol, charge, self_interaction = True):
    
    s1 = symbol
    z1 = atomic_numbers[symbol]
    q1 = charge
    n1 = principle_number[z1]
    rc1 = covalent_radii[z1]
    r0, r_min, alpha, d0, n2, rc2, mask = [], [], [], [], [], [], []
    for i, (s2, q2, z2) in enumerate(zip(atoms.symbols, atoms.get_array('oxi_states'), atoms.numbers)):
        if q2 < 0:
            params = bvse_data[s1][q1][s2][q2]
            r0.append(params['r0']) # dont need it, actually
            r_min.append(params['r_min'])
            alpha.append(params['alpha'])
            d0.append(params['d0'])
            n2.append(principle_number[z2]); rc2.append(covalent_radii[z2])
            mask.append(1.0)
        else:
            r0.append(0); r_min.append(0); alpha.append(0); d0.append(0)
            n2.append(principle_number[z2])
            rc2.append(covalent_radii[z2])
            if z1 == z2:
                if self_interaction:
                    mask.append(1.0)
                else:
                    mask.append(0.0)
            else:
                mask.append(1.0)
    coulomb_params = {'n2': n2, 'rc2': rc2, 'mask': mask}
    morse_params = {'r0': r0, 'd0': d0, 'alpha': alpha, 'r_min': r_min}
    for key in morse_params:
        atoms.set_array(key, np.array(morse_params[key]))
    for key in coulomb_params:
        atoms.set_array(key, np.array(coulomb_params[key]))
    return atoms



def geometric_median(X, y0, eps=1e-5, max_iter = 200):
    """
    Finds geometric median of the X points using y0 as the initial guess
    """
    #y = np.mean(X, 0)
    y = y0
    counter = 0
    while True:
        counter += 1
        D = cdist(X, [y])
        nonzeros = (D != 0)[:, 0]

        Dinv = 1 / D[nonzeros]
        Dinvs = np.sum(Dinv)
        W = Dinv / Dinvs
        T = np.sum(W * X[nonzeros], 0)

        num_zeros = len(X) - np.sum(nonzeros)
        if num_zeros == 0:
            y1 = T
        elif num_zeros == len(X):
            return y
        else:
            R = (T - y) * Dinvs
            r = np.linalg.norm(R)
            rinv = 0 if r == 0 else num_zeros/r
            y1 = max(0, 1-rinv)*T + min(1, rinv)*y

        if euclidean(y, y1) < eps:
            return y1
        
        
        if counter > max_iter:
            return y
        y = y1


def geometric_median_string(left, center, right, source):

    base = np.zeros((len(left)+1, 3))
    base[:len(left), :] = left.positions
    base[-1, :] = right.positions[source]
    translations = np.array(list(itertools.product(
                                    [0, 1, -1], # should be [0, 1, -1, 2, -2] ideally
                                    [0, 1, -1],
                                    [0, 1, -1])))
    coords = []
    for tr in translations:
        coords.append(base + np.dot(tr, left.cell))
    X = np.vstack(coords)
    y0 = center.positions[source]
    gm = geometric_median(X, y0)
    return gm, y0





def geometric_median_with_bounds(left, center, right, source):

    """bounds are hard coded"""

    base = np.zeros((len(left)+1, 3))
    base[:len(left), :] = left.copy().positions
    base[-1, :] = right.copy().positions[source]
    translations = np.array(list(itertools.product(
                                    [0, 1, -1], # should be [0, 1, -1, 2, -2] ideally
                                    [0, 1, -1],
                                    [0, 1, -1])))
    coords = []
    for tr in translations:
        coords.append(base + np.dot(tr, left.copy().cell))
    X = np.vstack(coords)

    x = X[:, 0]
    y = X[:, 1]
    z = X[:, 2]
    x0 = center.copy().positions[source]
    def dist_func(x0):
        return sum(((np.full(len(x),x0[0])-x)**2+(np.full(len(x),x0[1])-y)**2+(np.full(len(x),x0[2])-z)**2)**(1/2))
    
    bounds = ((x0[0] - 0.25, x0[0] + 0.25),(x0[1] - 0.25, x0[1] + 0.25), (x0[2] - 0.25, x0[2] + 0.25))
    res = minimize(dist_func, x0, method='nelder-mead', options={'xtol': 1e-8, 'disp': False}, bounds=bounds)
    return res.x, x0